package ru.yandex.direct.core.entity.bidmodifiers.validation.typesupport;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;

import org.springframework.stereotype.Component;

import ru.yandex.direct.core.entity.adgroup.model.AdGroup;
import ru.yandex.direct.core.entity.bidmodifier.BidModifier;
import ru.yandex.direct.core.entity.bidmodifier.BidModifierSmartTV;
import ru.yandex.direct.core.entity.bidmodifier.BidModifierType;
import ru.yandex.direct.core.entity.bidmodifiers.container.BidModifierKey;
import ru.yandex.direct.core.entity.bidmodifiers.service.CachingFeaturesProvider;
import ru.yandex.direct.core.entity.campaign.model.CampaignType;
import ru.yandex.direct.dbutil.model.ClientId;
import ru.yandex.direct.validation.builder.ListValidationBuilder;
import ru.yandex.direct.validation.result.Defect;
import ru.yandex.direct.validation.result.ValidationResult;
import ru.yandex.direct.validation.wrapper.ModelItemValidationBuilder;

import static ru.yandex.direct.core.entity.bidmodifier.BidModifierSmartTV.SMART_T_V_ADJUSTMENT;
import static ru.yandex.direct.core.entity.bidmodifier.BidModifierType.SMARTTV_MULTIPLIER;
import static ru.yandex.direct.core.entity.bidmodifiers.validation.typesupport.BidModifierValidationHelper.validateAdjustmentCommon;
import static ru.yandex.direct.core.entity.bidmodifiers.validation.typesupport.DeviceModifiersConflictChecker.setDeviceBidModifierAllZerosDefect;
import static ru.yandex.direct.core.entity.bidmodifiers.validation.typesupport.DeviceModifiersConflictChecker.validateModifierDoNotRewriteExisting;

@Component
@ParametersAreNonnullByDefault
public class BidModifierValidationSmartTVTypeSupport implements BidModifierValidationTypeSupport<BidModifierSmartTV> {
    private final DeviceModifiersConflictChecker deviceModifiersConflictChecker;

    public static boolean isSmartTvEnabled(CampaignType type) {
        return type.equals(CampaignType.CPM_BANNER) || type.equals(CampaignType.CPM_DEALS);
    }
    public static Predicate<BidModifier> getZeroValuePredicate() {
        return m -> ((BidModifierSmartTV) m).getSmartTVAdjustment().getPercent().equals(0);
    }

    public static void fillZeroValuePredicate(Map<BidModifierType, Predicate<BidModifier>> typeRules) {
        typeRules.put(SMARTTV_MULTIPLIER, getZeroValuePredicate());
    }

    public static Set<Map<BidModifierType, Predicate<BidModifier>>> getAdjacentRules() {
        var rules1 = new HashMap<BidModifierType, Predicate<BidModifier>>();
        BidModifierValidationDesktopTypeSupport.fillZeroValuePredicate(rules1);
        BidModifierValidationMobileTypeSupport.fillZeroValuePredicate(rules1);

        var rules2 = new HashMap<BidModifierType, Predicate<BidModifier>>();
        BidModifierValidationDesktopOnlyTypeSupport.fillZeroValuePredicate(rules2);
        BidModifierValidationTabletTypeSupport.fillZeroValuePredicate(rules2);
        BidModifierValidationMobileTypeSupport.fillZeroValuePredicate(rules2);

        return Set.of(rules1, rules2);
    }

    public BidModifierValidationSmartTVTypeSupport(DeviceModifiersConflictChecker deviceModifiersConflictChecker) {
        this.deviceModifiersConflictChecker = deviceModifiersConflictChecker;
    }

    @Override
    public BidModifierType getType() {
        return BidModifierType.SMARTTV_MULTIPLIER;
    }

    @Override
    public ValidationResult<BidModifierSmartTV, Defect> validateAddStep1(BidModifierSmartTV modifier,
                                                                         CampaignType campaignType,
                                                                         @Nullable AdGroup adGroupWithType, ClientId clientId,
                                                                         CachingFeaturesProvider featuresProvider) {
        ModelItemValidationBuilder<BidModifierSmartTV> vb = ModelItemValidationBuilder.of(modifier);
        vb.item(BidModifierSmartTV.SMART_T_V_ADJUSTMENT)
                .checkBy(adjustment -> validateAdjustmentCommon(adjustment, getType(), campaignType,
                        adGroupWithType, clientId, featuresProvider));
        return vb.getResult();
    }

    @Override
    public ValidationResult<List<BidModifierSmartTV>, Defect> validateAddStep2(
            ClientId clientId,
            List<BidModifierSmartTV> modifiersToValidate, Map<BidModifierKey, BidModifierSmartTV> existingModifiers,
            Map<Long, CampaignType> campaignTypes, Map<Long, AdGroup> adGroupsWithType,
            CachingFeaturesProvider featuresProvider,
            Map<BidModifierKey, BidModifier> allValidBidModifiersInOperation) {

        ListValidationBuilder<BidModifierSmartTV, Defect> lvb = ListValidationBuilder.of(modifiersToValidate);

        lvb.checkEachBy(modifier -> validateModifierDoNotRewriteExisting(
                modifier, existingModifiers, SMART_T_V_ADJUSTMENT));

        var mediaModifiers = modifiersToValidate.stream()
                .filter(m -> isSmartTvEnabled(campaignTypes.get(m.getCampaignId())))
                .collect(Collectors.toList());

        Set<BidModifierKey> zeroValueConflicts = checkNoZeroValue(
                        clientId, mediaModifiers, allValidBidModifiersInOperation);

        lvb.checkEach(setDeviceBidModifierAllZerosDefect(zeroValueConflicts));

        return lvb.getResult();
    }

    private Set<BidModifierKey> checkNoZeroValue(
            ClientId clientId,
            List<BidModifierSmartTV> modifiersToValidate,
            Map<BidModifierKey, BidModifier> allValidBidModifiersInOperation) {

        Set<BidModifierKey> result = new HashSet<>();

        for (var rulesSet : getAdjacentRules()) {
            result.addAll(deviceModifiersConflictChecker.findConflictingModifiers(
                    clientId, modifiersToValidate, allValidBidModifiersInOperation,
                    k -> getZeroValuePredicate().test(k),
                    rulesSet
            ));
        }

        return result;
    }

}
