package ru.yandex.direct.grid.processing.service.group.validation;

import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;

import one.util.streamex.EntryStream;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import ru.yandex.direct.core.entity.adgroup.model.AdGroupType;
import ru.yandex.direct.core.entity.adgroup.model.PerformanceAdGroup;
import ru.yandex.direct.core.entity.adgroup.service.AdGroupService;
import ru.yandex.direct.dbutil.model.ClientId;
import ru.yandex.direct.grid.processing.model.api.GdValidationResult;
import ru.yandex.direct.grid.processing.model.group.mutation.GdAddSmartAdGroup;
import ru.yandex.direct.grid.processing.model.group.mutation.GdAddSmartAdGroupItem;
import ru.yandex.direct.grid.processing.model.group.mutation.GdUpdatePerformanceAdGroup;
import ru.yandex.direct.grid.processing.model.group.mutation.GdUpdatePerformanceAdGroupItem;
import ru.yandex.direct.grid.processing.service.validation.GridDefectDefinitions;
import ru.yandex.direct.grid.processing.service.validation.GridValidationResultConversionService;
import ru.yandex.direct.grid.processing.service.validation.GridValidationService;
import ru.yandex.direct.validation.builder.Validator;
import ru.yandex.direct.validation.builder.When;
import ru.yandex.direct.validation.result.DefaultPathNodeConverterProvider;
import ru.yandex.direct.validation.result.Defect;
import ru.yandex.direct.validation.result.Path;
import ru.yandex.direct.validation.result.PathNodeConverterProvider;
import ru.yandex.direct.validation.result.ValidationResult;
import ru.yandex.direct.validation.wrapper.ModelItemValidationBuilder;

import static ru.yandex.direct.grid.processing.service.validation.presentation.AdGroupConverters.ADD_PERFORMANCE_AD_GROUP_PATH_CONVERTER;
import static ru.yandex.direct.utils.FunctionalUtils.mapList;
import static ru.yandex.direct.validation.constraint.CollectionConstraints.notEmptyCollection;
import static ru.yandex.direct.validation.constraint.CommonConstraints.eachNotNull;
import static ru.yandex.direct.validation.constraint.CommonConstraints.inSet;
import static ru.yandex.direct.validation.constraint.CommonConstraints.notNull;
import static ru.yandex.direct.validation.constraint.CommonConstraints.validId;
import static ru.yandex.direct.validation.defect.CommonDefects.objectNotFound;

@Service
@ParametersAreNonnullByDefault
public class PerformanceAdGroupValidationService {
    private final GridValidationService gridValidationService;
    private final AdGroupService adGroupService;
    private final PathNodeConverterProvider pathNodeConverterProvider;

    @Autowired
    public PerformanceAdGroupValidationService(GridValidationService gridValidationService,
                                               AdGroupService adGroupService) {
        this.gridValidationService = gridValidationService;
        this.adGroupService = adGroupService;
        this.pathNodeConverterProvider = DefaultPathNodeConverterProvider.builder()
                .register(PerformanceAdGroup.class, ADD_PERFORMANCE_AD_GROUP_PATH_CONVERTER)
                .build();
    }

    public void validateUpdatePerformanceAdGroups(ClientId clientId, GdUpdatePerformanceAdGroup gdUpdateRequest) {
        gridValidationService.applyValidator(req -> validateUpdateRequest(req, clientId), gdUpdateRequest, false);
    }

    private ValidationResult<GdUpdatePerformanceAdGroup, Defect> validateUpdateRequest(GdUpdatePerformanceAdGroup req,
                                                                                       ClientId clientId) {
        List<Long> adGroupIds = mapList(req.getUpdateItems(), GdUpdatePerformanceAdGroupItem::getId);
        Map<Long, AdGroupType> adGroupTypes = adGroupService.getAdGroupTypes(clientId, adGroupIds);
        Set<Long> existAdGroupIds = adGroupTypes.keySet();
        Set<Long> performanceAdGroupIds = EntryStream.of(adGroupTypes)
                .filterValues(v -> v == AdGroupType.PERFORMANCE)
                .mapKeyValue((k, v) -> k)
                .toSet();
        ModelItemValidationBuilder<GdUpdatePerformanceAdGroup> vb = ModelItemValidationBuilder.of(req);
        vb.list(GdUpdatePerformanceAdGroup.UPDATE_ITEMS)
                .checkEachBy(item -> validateUpdateItem(item, existAdGroupIds, performanceAdGroupIds));
        return vb.getResult();
    }

    private ValidationResult<GdUpdatePerformanceAdGroupItem, Defect> validateUpdateItem(
            GdUpdatePerformanceAdGroupItem item,
            Set<Long> existAdGroupIds,
            Set<Long> performanceAdGroupIds) {
        ModelItemValidationBuilder<GdUpdatePerformanceAdGroupItem> vb = ModelItemValidationBuilder.of(item);
        vb.item(GdUpdatePerformanceAdGroupItem.ID)
                .check(validId())
                .check(inSet(existAdGroupIds), objectNotFound(), When.isValid())
                .check(inSet(performanceAdGroupIds), GridDefectDefinitions.unsupportedAdGroupType(), When.isValid());
        vb.item(GdUpdatePerformanceAdGroupItem.MINUS_KEYWORDS)
                .check(eachNotNull());
        vb.item(GdUpdatePerformanceAdGroupItem.REGION_IDS)
                .check(eachNotNull());
        return vb.getResult();
    }

    public void validateAddPerformanceAdGroups(GdAddSmartAdGroup gdAddRequest) {
        gridValidationService.applyValidator(AD_GROUP_ADD_VALIDATOR, gdAddRequest, false);
    }

    private static final Validator<GdAddSmartAdGroupItem, Defect> AD_GROUP_ADD_ITEMS_VALIDATOR = items -> {
        ModelItemValidationBuilder<GdAddSmartAdGroupItem> vb = ModelItemValidationBuilder.of(items);
        vb.item(GdAddSmartAdGroupItem.CAMPAIGN_ID)
                .check(validId(), When.isValid());
        vb.item(GdAddSmartAdGroupItem.NAME)
                .check(notNull());
        return vb.getResult();
    };

    private static final Validator<GdAddSmartAdGroup, Defect> AD_GROUP_ADD_VALIDATOR = gdAddSmartAdGroup -> {
        ModelItemValidationBuilder<GdAddSmartAdGroup> vb = ModelItemValidationBuilder.of(gdAddSmartAdGroup);
        vb.list(GdAddSmartAdGroup.ADD_ITEMS)
                .check(notEmptyCollection(), When.notNull())
                .checkEachBy(AD_GROUP_ADD_ITEMS_VALIDATOR);
        return vb.getResult();
    };

    @Nullable
    public GdValidationResult getValidationResult(ValidationResult<?, Defect> vr, Path path) {
        return GridValidationResultConversionService
                .buildGridValidationResultIfErrors(vr, path, pathNodeConverterProvider);
    }
}
